/****************************************************************************
 *
 *   Copyright (c) 2014 MAVlink Development Team. All rights reserved.
 *   Author: Trent Lukaczyk, <aerialhedgehog@gmail.com>
 *           Jaycee Lock,    <jaycee.lock@gmail.com>
 *           Lorenz Meier,   <lm@inf.ethz.ch>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name PX4 nor the names of its contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 ****************************************************************************/

/**
 * @file serial_port.cpp
 *
 * @brief Serial interface functions
 *
 * Functions for opening, closing, reading and writing via serial ports
 *
 * @author Trent Lukaczyk, <aerialhedgehog@gmail.com>
 * @author Jaycee Lock,    <jaycee.lock@gmail.com>
 * @author Lorenz Meier,   <lm@inf.ethz.ch>
 *
 */


// ------------------------------------------------------------------------------
//   Includes
// ------------------------------------------------------------------------------

#include "serial_port.h"
#include <px4_log.h>
#include <unistd.h>
#include <sys/ioctl.h>
// ----------------------------------------------------------------------------------
//   Serial Port Manager Class
// ----------------------------------------------------------------------------------

// ------------------------------------------------------------------------------
//   Con/De structors
// ------------------------------------------------------------------------------
SerialPort::SerialPort()
{
	// Initialize attributes
	fd     = -1;
	status = SERIAL_PORT_CLOSED;
}


SerialPort::~SerialPort()
{
}


// ------------------------------------------------------------------------------
//   Read from Serial
// ------------------------------------------------------------------------------


//Read exactly a frame of "len" number of bytes
void SerialPort::read_port_frame(char *buf, int len, bool showPrints)
{
	int bytesReadTotal=0;
	int bytesRead=0;

	while(bytesReadTotal != len){
		bytesRead = read(fd, &buf[bytesReadTotal], (unsigned)len);
		bytesReadTotal += bytesRead;

	}

	int num =0;
	if (bytesRead == -1)

	{
	if (showPrints)	fprintf(stderr,"\nERROR: could not read port\n");

	}else{

		if (showPrints){

	  	printf("\nMessage read correctly, bytes read in total: %d\n", bytesReadTotal);
		  for(int i=0; i<bytesReadTotal; i++){
				if((int8_t)buf[i] < 0)
					num = (int8_t)buf[i] + 256;
				else
					num = (int8_t)buf[i];

	       			printf("I(%d): %d\n",i,num);
				  }
			  }
		}
	//return buf;
}

//Synchronize communication to let the frames read correctly
int SerialPort::resynchronize(char *buf)
{

		int bytesCorrect=0, i=0;
	  bool stxFound=false;
		//Search STX in the incorrect buffer
		while(!stxFound){
				if (buf[i] == STX){
					  buf[0]=buf[i];
						bytesCorrect++;
						stxFound = true;
				}
				i++;
		}

		//Save the correct part of the buffer
		while(i<16){
				buf[bytesCorrect] = buf[i];
				bytesCorrect++;
				i++;
		}


 	return bytesCorrect;
}

// ------------------------------------------------------------------------------
//   Write to Serial
// ------------------------------------------------------------------------------


int SerialPort::send_message_buf(char* buf, int len)
{
    // Write buffer to serial port, locks port while writing
    const int bytesWritten = write(fd, (char*) buf, (unsigned)len);

    return bytesWritten;
}


// ------------------------------------------------------------------------------
//   Open Serial Port
// ------------------------------------------------------------------------------
/**
 * throws EXIT_FAILURE if could not open the port
 */
void SerialPort::open_serial()
{
    printf("OPEN PORT: %s\n", m_uart_name);

    fd = _open_port();

    if (fd == -1)
    {
        printf("failure, could not open port.\n");
        return;
    }

    bool success = _setup_port(m_baudRate, 8, 1, false, false);

    if (!success)
    {
        printf("Failure, could not configure port.\n");
        return;
    }
    if (fd <= 0)
    {
        printf("Connection attempt to port %s with %d baud, 8N1 failed, exiting.\n", m_uart_name, m_baudRate);
        return;
    }

    printf("Connected to %s with %d baud, 8 data bits, no parity, 1 stop bit (8N1)\n", m_uart_name, m_baudRate);

    status = true;

    printf("\n");
}


// ------------------------------------------------------------------------------
//   Close Serial Port
// ------------------------------------------------------------------------------
void SerialPort::close_serial()
{
    printf("CLOSE PORT\n");

    int result = close(fd);

    if ( result )
    {
        fprintf(stderr,"WARNING: Error on port close (%i)\n", result );
    }

    status = false;

    printf("\n");
}


// ------------------------------------------------------------------------------
//   Convenience Functions
// ------------------------------------------------------------------------------
void SerialPort::start( const char* portName, int baudRate, bool useNonBlockingIO, bool debug)
{
    m_uart_name= portName;
    m_baudRate = baudRate;

    m_numBytesRead= 0;
    m_magicNumberRead = false;
    m_useNonBlockingIO= useNonBlockingIO;
    m_debug= debug;

    open_serial();
}

void SerialPort::stop()
{
    close_serial();
}


// ------------------------------------------------------------------------------
//   Quit Handler
// ------------------------------------------------------------------------------
void SerialPort::handle_quit( int sig )
{
    stop();
}


// ------------------------------------------------------------------------------
//   Helper Function - Open Serial Port File Descriptor
// ------------------------------------------------------------------------------
// Where the actual port opening happens, returns file descriptor 'fd'
int SerialPort::_open_port()
{
    // Open serial port
    // O_NOCTTY - Ignore special chars like CTRL-C
    if (!m_useNonBlockingIO)
        fd = open(m_uart_name, O_RDWR | O_NOCTTY );
    else
        fd = open(m_uart_name, O_RDWR | O_NOCTTY | O_NONBLOCK | O_NDELAY);

    // Check for Errors
    if (fd == -1)
    {
        fprintf(stderr, "\nERROR: could not open serial port\n");
        return(-1);
    }
    if (!m_useNonBlockingIO)
       fcntl(fd, F_SETFL, 0); //it was originally set as a blocking file descriptor

    return fd;
}

// ------------------------------------------------------------------------------
//   Helper Function - Setup Serial Port
// ------------------------------------------------------------------------------
// Sets configuration, flags, and baud rate
bool SerialPort::_setup_port(int baud, int data_bits, int stop_bits, bool parity, bool hardware_control)
{
	// Check file descriptor
	if(!isatty(fd))
	{
		fprintf(stderr, "\nERROR: file descriptor %d is NOT a serial port\n", fd);
		return false;
	}

	// Read file descritor configuration
	struct termios  config;

	if(tcgetattr(fd, &config) < 0)
	{
            printf("Error: could not retrieve port configuration (descriptor %d)\n", fd);
            return false;
        }

	// Apply baudrate
	switch (baud)
	{
	case 1200:
		if (cfsetispeed(&config, B1200) < 0 || cfsetospeed(&config, B1200) < 0)
		{
			fprintf(stderr, "\nERROR: Could not set desired baud rate of %d Baud\n", baud);
			return false;
		}
		break;
	case 1800:
		cfsetispeed(&config, B1800);
		cfsetospeed(&config, B1800);
		break;
	case 9600:
		cfsetispeed(&config, B9600);
		cfsetospeed(&config, B9600);
		break;
	case 19200:
		cfsetispeed(&config, B19200);
		cfsetospeed(&config, B19200);
		break;
	case 38400:
		if (cfsetispeed(&config, B38400) < 0 || cfsetospeed(&config, B38400) < 0)
		{
			fprintf(stderr, "\nERROR: Could not set desired baud rate of %d Baud\n", baud);
			return false;
		}
		break;
	case 57600:
		if (cfsetispeed(&config, B57600) < 0 || cfsetospeed(&config, B57600) < 0)
		{
			fprintf(stderr, "\nERROR: Could not set desired baud rate of %d Baud\n", baud);
			return false;
		}
		break;
	case 115200:
		if (cfsetispeed(&config, B115200) < 0 || cfsetospeed(&config, B115200) < 0)
		{
			fprintf(stderr, "\nERROR: Could not set desired baud rate of %d Baud\n", baud);
			return false;
		}
		break;

	// These two non-standard (by the 70'ties ) rates are fully supported on
	// current Debian and Mac OS versions (tested since 2010).
	case 460800:
		if (cfsetispeed(&config, B460800) < 0 || cfsetospeed(&config, B460800) < 0)
		{
			fprintf(stderr, "\nERROR: Could not set desired baud rate of %d Baud\n", baud);
			return false;
		}
		break;
	case 921600:
		if (cfsetispeed(&config, B921600) < 0 || cfsetospeed(&config, B921600) < 0)
		{
			fprintf(stderr, "\nERROR: Could not set desired baud rate of %d Baud\n", baud);
			return false;
		}
		break;
	default:
		fprintf(stderr, "ERROR: Desired baud rate %d could not be set, aborting.\n", baud);
		return false;

		break;
	}

        //disable the flow control??: UART/I2C port doesn't support it
	//if (m_useNonBlockingIO)
            config.c_cflag &= ~CRTSCTS;

	// Finally, apply the configuration
	if(tcsetattr(fd, TCSANOW, &config) < 0)
	{
		fprintf(stderr, "\nERROR: could not set configuration of fd %d\n", fd);
                perror("tcsetattr() error");

		return false;
	}
        printf("Input speed = %d // Output speed = %d\n", (int) cfgetispeed(&config), (int) cfgetospeed(&config));
        printf("Input flag = %d // Output flag = %d\n", (int) config.c_iflag, (int) config.c_oflag);
        printf("Control mode = %d // Local mode = %d\n", (int) config.c_cflag, (int) config.c_lflag);


	// Done!
	return true;
}
