#include "slave_controller.h"
#include "codisava.h"

#include <math.h>
#include <memory.h>

#include <px4_posix.h>
#include <px4_log.h>


#include <uORB/uORB.h>
#include <uORB/topics/manual_control_setpoint.h>
#include <uORB/topics/vehicle_attitude.h>

//Utility functions: quaternion -> eulerian angles
//http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToEuler/index.htm

//The reference axis seems to be different between px4 and the typical one used in 3d graphics, so I swapped yaw and pitch to make it work
double pitchFromQuaternion(float* quat)
{
	// pitch (y-axis rotation)
	double test = quat[0]*quat[1] + quat[2]*quat[3];
	if (test > 0.499) // singularity at north pole
		return 2 * atan2((double)quat[0], (double)quat[3]);

	if (test < -0.499) // singularity at south pole
		return -2 * atan2((double)quat[0], (double)quat[3]);

	double sqy = quat[1]*quat[1];
	double sqz = quat[2]*quat[2];
	return atan2((double) (2 * quat[1]*quat[3] - 2 * quat[0]*quat[2]), (double) (1 - 2 * sqy - 2 * sqz));
}

double yawFromQuaternion(float* quat)
{
	// yaw (x-axis rotation)
	double test = quat[0]*quat[1] + quat[2]*quat[3];
	if (test > 0.499) // singularity at north pole
		return 0.0;

	if (test < -0.499) // singularity at south pole
		return 0.0;

	double sqx = quat[0]*quat[0];
	double sqz = quat[2]*quat[2];

	return atan2((double) (2 * quat[0]*quat[3] - 2 * quat[1]*quat[2]), (double)( 1 - 2 * sqx - 2 * sqz));
}

double rollFromQuaternion(float* quat)
{
	// roll (z-axis rotation)
	double test = quat[0]*quat[1] + quat[2]*quat[3];
	if (test > 0.499) // singularity at north pole
		return M_PI * 0.5;

	if (test < -0.499) // singularity at south pole
		return -M_PI * 0.5;

		//PX4_INFO("roll lf = %lf", asin((double) (2 * test)));
	//	PX4_INFO("roll f = %f", asin((double) (2 * test)));
	return asin((double) (2 * test));
}


unsigned char CalculateChecksumB(char* buf)
{
    unsigned char checksum = 0;

    for (int byte= 1; byte<=13; byte++)
    {
        checksum = checksum xor buf[byte];
    }

    return checksum;
}



//This function converts the euler angle got with the lectures of the sensor to a 32 bytes angle
int convertProcessValue(double eu_angle)
{

	double angle = ((eu_angle + M_PI) * pow(2,31))/(2*M_PI);
	return (int)angle;
}

//This function is used to split the angle in a 4 bytes format
void split4bytes(int angle, unsigned char* angle4bytes)
{

	 //unsigned char angle4bytes[4];
	 int rest1, rest2, rest3;
	 int byte1, byte2, byte3, byte4;
	  //first byte
	 rest1=angle%16777216;
	 byte1=angle/16777216;
	  //second byte
	 rest2=rest1%65536;
	 byte2=rest1/65536;
	  //third byte
	 rest3=rest2%256;
	 byte3=rest2/256;
	  //fourth byte
	 byte4=rest3;
	// PX4_INFO("byte1 value: %d", byte1);
	 //PX4_INFO("byte2 value: %d", byte2);
	 //PX4_INFO("byte3 value: %d", byte3);
	// PX4_INFO("byte4 value: %d", byte4);

	 //Put the bytes in the vector
	 angle4bytes[0] = (unsigned char)byte1;
	 angle4bytes[1] = (unsigned char)byte2;
	 angle4bytes[2] = (unsigned char)byte3;
	 angle4bytes[3] = (unsigned char)byte4;

	 //return angle4bytes;
}

void bytesToAngles(double pitch, double roll, double thrust, char* buf)
{
			unsigned char pitch1 = buf[2];
			unsigned char pitch2 = buf[3];
			unsigned char pitch3 = buf[4];
			unsigned char pitch4 = buf[5];

			unsigned char roll1 = buf[6];
	    unsigned char roll2 = buf[7];
	    unsigned char roll3 = buf[8];
	    unsigned char roll4 = buf[9];

			unsigned char thrust1 = buf[10];
	    unsigned char thrust2 = buf[11];
	    unsigned char thrust3 = buf[12];
	    unsigned char thrust4 = buf[13];

	    unsigned long txPitch =pitch4 + pitch3*256 + pitch2*65536 + pitch1*16777216;
			unsigned long txRoll = roll4 + roll3*256 + roll2*65536 + roll1*16777216;
			unsigned long txThrust = thrust4 + thrust3*256 + thrust2*65536 + thrust1*16777216;


			bool show=false;


			pitch = (txPitch*M_PI)/pow(2,30) - M_PI;
			roll = (txRoll*M_PI)/pow(2,30) - M_PI;
			thrust = (txThrust)/pow(2,30);

     if(show){
			 PX4_INFO("Received pitch: %f", pitch);
			 PX4_INFO("Received Roll: %f", roll);
			 PX4_INFO("Received Thrust: %f", thrust);
		 }



}





SlaveController::SlaveController()
{
}

SlaveController::~SlaveController()
{
}

void SlaveController::ParseArguments(int argc, char* argv[])
{
    for (int i= 0; i<argc; i++)
    {
        sscanf(argv[i],"id=%d",&m_id);
        if (!strcmp(argv[i],"no-feedback")) m_sendFeedback= false;
        if (!strcmp(argv[i],"debug")) m_outputConsoleMsgs= true;
        if (!strcmp(argv[i],"non-blocking")) m_useNonBlockingIO = true;
    }
}

bool SlaveController::IsValid(char* buf)
{
    //first, the easy part: check the beginning and the end of the message
    if (buf[0] != STX || buf[15] != ETX)
		{
		//	PX4_INFO("STX and ETX are incorrect");
			return false;
		}
    //now, the checksum
    if (CalculateChecksumB(buf) == buf[14])
		{
			//PX4_INFO("CORRECT MESSAGE");
    	return true;
		}
		//PX4_INFO("INVALID CHECKSUM");
    return false;
}

int SlaveController::Run(int argc, char* argv[])
{


    //parse the controller's arguments
    ParseArguments(argc, argv);

    if (m_id<0 || m_id>4)
    {
        PX4_INFO("Error: invalid id. It must be an integer number between 1 and 4");
        return -1;
    }

    PX4_INFO("Execution mode: One-way communication slave controller\nID=%d Feedback=%d Debug=%d Non-blocking=%d"
         , m_id, m_sendFeedback, m_outputConsoleMsgs, m_useNonBlockingIO);

    //Initialize serial ports: input and output
    m_inputPort.start(CODISAVA_INPUT_PORT, CODISAVA_BAUD_RATE, m_useNonBlockingIO, m_outputConsoleMsgs); //use m_outputConsoleMsgs instead of false for more serial-port info messages
    m_outputPort.start(CODISAVA_OUTPUT_PORT, CODISAVA_BAUD_RATE, m_useNonBlockingIO, m_outputConsoleMsgs); //use m_outputConsoleMsgs instead of false for more serial-port info messages

    //Input -> Output/controller initialization (main thread)
    //advertise actuator_control topic so that we can send commands to the local actuator
    struct manual_control_setpoint_s act;
    memset(&act, 0, sizeof(act));
    orb_advert_t act_pub = orb_advertise(ORB_ID(manual_control_setpoint), &act);

		//advertise topic to arm the drone
		struct actuator_armed_s arm;
    memset(&arm, 0, sizeof(arm));
    orb_advert_t act_pub_armed = orb_advertise(ORB_ID(actuator_armed), &arm);


    //Sensors used for feedback initialization
    int sensor_sub_fd = orb_subscribe(ORB_ID(vehicle_attitude));    // subscribe to sensor_combined topic
    orb_set_interval(sensor_sub_fd, CODISAVA_SENSOR_READING_FREQ);     // limit the update rate

    px4_pollfd_struct_t sensor_fd;
    sensor_fd.fd = sensor_sub_fd;
    sensor_fd.events = POLLIN;



    int error_counter = 0;


		//Variables to save sensor data
		double rxPitch= 0, rxRoll = 0, rxThrust=0;
    double lastPitch= 0, lastRoll = 0;
		int truePitch= 0, trueRoll = 0;

		//buffers to save pitch, roll and thrust in 4 bytes
		unsigned char pitch4bytes[4];
		unsigned char roll4bytes[4]; //throttle4bytes[4];
    //int read_ret= 0;
		//Main fuffer
		char buf[16];
		//Initialize buffers
		for(int i=0; i<16; i++)  buf[i] = 0;
		//Variable for the written bytes to the master
		int bytesWritten;
		//Variables to control the synchro of the communication
		int incorrectCount = 0, syncBytes=0;
		int correctFrames =0, framesSent=0, framesReceived=0;

		//variables to control the monitorizing freq
		int showPrintsCount = 0;
		bool showPrints;
		//Start the main loop
    while (true)
    {

			showPrintsCount++;
			//Show the prints
			if (showPrintsCount > 1000){
				showPrints = true;
				showPrintsCount = 0;
			} else if(showPrints) showPrints=false;
			//Recieve the starting message
		  m_inputPort.read_port_frame(&buf[syncBytes], 16-syncBytes, showPrints);
			framesReceived++;

			if (showPrints) PX4_INFO("Checksum value = %d", (int)CalculateChecksumB(buf));
			//If the message id valid, publish the data in actuators
			if(IsValid(buf)){
					syncBytes=0;
					//publish received command for the actuator
					correctFrames++;
					if (showPrints) PX4_INFO("CORRECT MESSAGE");

					//Take buffer values and convert them to angles
					bytesToAngles(rxPitch, rxRoll, rxThrust, buf);
					act.x = rxRoll/M_PI;    // roll
					act.y = rxPitch/M_PI;   // pitch
					act.z = rxThrust;			 // thrust
					act.r = 0;     		 // yaw

					//Arm or disarm depending on the ID byte
					if((buf[1] & (1<<1)) == 1)
				  	arm.armed=true;
					else
						arm.armed=false

					orb_publish(ORB_ID(actuator_armed), act_pub_armed, &arm);
					orb_publish(ORB_ID(manual_control_setpoint), act_pub, &act);


		 }else{
			 		//If lot of frames incorrect, then try to synchronize
			 		incorrectCount++;
					if(incorrectCount >= 1){
							syncBytes = m_inputPort.resynchronize(buf);
							incorrectCount=0;
					}
		 }

			//Read sensor values and adapt them to send
		 int poll_ret = px4_poll(&sensor_fd, 1, 100); //we wait up to 0.1 secs
 		 if (poll_ret < 0)
 		 {
 				 // this is seriously bad - should be an emergency
 				 if (error_counter < 10 || error_counter % 50 == 0)
 				 {
 						 PX4_ERR("ERROR return value from poll(): %d", poll_ret);
 				 }
 				 error_counter++;
 		 }
 		 else if (poll_ret>0 && sensor_fd.revents & POLLIN)
 		 {
 				 //there is sensor data available: get it and dispatch it to the previous slave or master
 				 struct vehicle_attitude_s raw_sensor;
 				 orb_copy(ORB_ID(vehicle_attitude), sensor_sub_fd, &raw_sensor);

 				 //lastYaw= yawFromQuaternion(raw_sensor.q);
				 //PX4_INFO("Raw sensor value: %f, %f, %f, %f", (double)raw_sensor.q[0], (double)raw_sensor.q[1],(double)raw_sensor.q[2], (double)raw_sensor.q[3]);
 				 lastPitch= pitchFromQuaternion(raw_sensor.q);
			   //PX4_INFO("Pitch value from sensor: %lf", lastPitch);
 				 lastRoll= rollFromQuaternion(raw_sensor.q);
				 //PX4_INFO("Roll value from sensor: %lf", lastRoll);
				 //Get angle value
				 truePitch = convertProcessValue(lastPitch);
				 // PX4_INFO("True pitch value: %d", truePitch);
				 trueRoll = convertProcessValue(lastRoll);
				 //PX4_INFO("True roll value: %d", trueRoll);
				 //Split in 4 bytes
				 split4bytes(truePitch, pitch4bytes);
				 split4bytes(trueRoll, roll4bytes);

 		 }

		 //Set the frame ready to send
		 buf[0] = 2;
		 buf[1] = 0;
		 buf[2] = pitch4bytes[0];
		 buf[3] = pitch4bytes[1];
		 buf[4] = pitch4bytes[2];
		 buf[5] = pitch4bytes[3];
		 buf[6] = roll4bytes[0];
		 buf[7] = roll4bytes[1];
		 buf[8] = roll4bytes[2];
		 buf[9] = roll4bytes[3];
		 buf[10] = 0;
		 buf[11] = 0;
		 buf[12] = 0;
		 buf[13] = 0;
		 buf[14] = CalculateChecksumB(buf);
		 buf[15] = 3;

		 //Send the frame to the master
		 bytesWritten = m_outputPort.send_message_buf(buf, 16);
		 framesSent++;
	   if (showPrints) PX4_INFO("Message sent, bytes written: %d", bytesWritten);
	   if (showPrints) PX4_INFO("Frames SENT: %d", framesSent);
		 if (showPrints) PX4_INFO("Frames RECEIVED: %d", framesReceived);
		 if (showPrints) PX4_INFO("Frames CORRECT: %d", correctFrames);
		 usleep(1);

    }

    return 0;
}
